@@ -2,13 +2,13 @@ use std::fmt;
 use std::str::FromStr;
 
 use async_channel::Sender;
-use lazy_static::lazy_static;
 use log::{debug, error, info};
 use rumqttc::{
     ConnAck, Connect, Event, MqttOptions, Packet, PubAck, PubComp, PubRec, PubRel, Publish,
     Request, SubAck, Subscribe, UnsubAck, Unsubscribe,
 };
 use serde::{Deserialize, Serialize};
+use tokio::sync::mpsc::UnboundedReceiver;
 
 use super::*;
 use crate::{AccountId, Addressable, AgentId, Authenticable, Error, SharedGroup};
@@ -18,25 +18,6 @@ use crate::queue_counter::QueueCounterHandle;
 
 const DEFAULT_MQTT_REQUESTS_CHAN_SIZE: Option<usize> = Some(10_000);
 
-lazy_static! {
-    static ref TOKIO: tokio::runtime::Runtime = {
-        let mut rt_builder = tokio::runtime::Builder::new_multi_thread();
-        rt_builder.enable_all();
-
-        let thread_count = std::env::var("TOKIO_THREAD_COUNT").ok().map(|value| {
-            value
-                .parse::<usize>()
-                .expect("Error converting TOKIO_THREAD_COUNT variable into usize")
-        });
-
-        if let Some(value) = thread_count {
-            rt_builder.worker_threads(value);
-        }
-
-        rt_builder.build().expect("Failed to start tokio runtime")
-    };
-}
-
 ////////////////////////////////////////////////////////////////////////////////
 
 /// Agent configuration.
@@ -165,40 +146,35 @@ impl AgentBuilder {
     ///     }
     /// }
     /// ```
-    pub fn start(
+    pub async fn start(
         self,
         config: &AgentConfig,
-    ) -> Result<(Agent, crossbeam_channel::Receiver<AgentNotification>), Error> {
-        self.start_with_runtime(config, TOKIO.handle().clone())
-    }
-
-    pub fn start_with_runtime(
-        self,
-        config: &AgentConfig,
-        rt_handle: tokio::runtime::Handle,
-    ) -> Result<(Agent, crossbeam_channel::Receiver<AgentNotification>), Error> {
-        let options = Self::mqtt_options(&self.connection, &config)?;
-        let channel_size = config
-            .requests_channel_size
-            .expect("requests_channel_size is not specified");
-        let mut eventloop = rumqttc::EventLoop::new(options, channel_size);
-        let mqtt_tx = eventloop.handle();
-        let reconnect_interval = config.reconnect_interval.to_owned();
-        let (tx, rx) = crossbeam_channel::unbounded::<AgentNotification>();
-        #[cfg(feature = "queue-counter")]
-        let queue_counter = QueueCounterHandle::start();
-        #[cfg(feature = "queue-counter")]
-        let queue_counter_ = queue_counter.clone();
-
-        std::thread::Builder::new()
-            .name("svc-agent-notifications-loop".to_owned())
-            .spawn(move || {
-                #[cfg(feature = "queue-counter")]
-                let queue_counter_ = queue_counter_.clone();
-                rt_handle.block_on(async {
-                    loop {
-                        match eventloop.poll().await {
-                            Ok(packet) => match packet {
+    ) -> Result<(Agent, UnboundedReceiver<AgentNotification>), Error> {
+        {
+            let options = Self::mqtt_options(&self.connection, config)?;
+            let channel_size = config
+                .requests_channel_size
+                .expect("requests_channel_size is not specified");
+            let mut eventloop = rumqttc::EventLoop::new(options, channel_size);
+            let mqtt_tx = eventloop.handle();
+            let reconnect_interval = config.reconnect_interval.to_owned();
+            let (tx, rx) = tokio::sync::mpsc::unbounded_channel::<AgentNotification>();
+            #[cfg(feature = "queue-counter")]
+            let queue_counter = QueueCounterHandle::start();
+            #[cfg(feature = "queue-counter")]
+            let queue_counter_ = queue_counter.clone();
+            tokio::spawn(async move {
+                let mut recovering_connection = false;
+                loop {
+                    match eventloop.poll().await {
+                        Ok(packet) => {
+                            if recovering_connection {
+                                recovering_connection = false;
+                                if let Err(e) = tx.send(AgentNotification::Reconnection) {
+                                    error!("Failed to notify about reconnection: {}", e);
+                                }
+                            }
+                            match packet {
                                 Event::Outgoing(content) => {
                                     info!("Outgoing message = '{:?}'", content);
                                 }
@@ -214,41 +190,39 @@ impl AgentBuilder {
                                         #[cfg(feature = "queue-counter")]
                                         queue_counter_.add_incoming_message(content);
                                     }
+
                                     if let Err(e) = tx.send(msg) {
                                         error!("Failed to transmit message, reason = {}", e);
                                     };
                                 }
-                            },
-                            Err(err) => {
-                                error!("Failed to poll, reason = {}", err);
-
-                                if let Err(e) = tx.send(AgentNotification::Disconnection) {
-                                    error!("Failed to notify about disconnection: {}", e);
-                                }
-                                match reconnect_interval {
-                                    Some(value) => {
-                                        tokio::time::sleep(std::time::Duration::from_secs(value))
-                                            .await
-                                    }
-                                    None => break,
+                            }
+                        }
+                        Err(err) => {
+                            error!("Failed to poll, reason = {}", err);
+                            recovering_connection = true;
+                            if let Err(e) = tx.send(AgentNotification::ConnectionError) {
+                                error!("Failed to notify about connection error: {}", e);
+                            }
+                            match reconnect_interval {
+                                Some(value) => {
+                                    tokio::time::sleep(std::time::Duration::from_secs(value)).await
                                 }
+                                None => break,
                             }
                         }
                     }
-                });
-            })
-            .map_err(|e| {
-                Error::new(&format!("Failed starting notifications loop thread, {}", e))
-            })?;
-        let agent = Agent::new(
-            self.connection.agent_id,
-            &self.api_version,
-            mqtt_tx,
-            #[cfg(feature = "queue-counter")]
-            queue_counter,
-        );
+                }
+            });
+            let agent = Agent::new(
+                self.connection.agent_id,
+                &self.api_version,
+                mqtt_tx,
+                #[cfg(feature = "queue-counter")]
+                queue_counter,
+            );
 
-        Ok((agent, rx))
+            Ok((agent, rx))
+        }
     }
 
     fn mqtt_options(connection: &Connection, config: &AgentConfig) -> Result<MqttOptions, Error> {
@@ -257,7 +231,9 @@ impl AgentBuilder {
             .parse::<http::Uri>()
             .map_err(|e| Error::new(&format!("error parsing MQTT connection URL, {}", e)))?;
         let host = uri.host().ok_or_else(|| Error::new("missing MQTT host"))?;
-        let port = uri.port().ok_or_else(|| Error::new("missing MQTT port"))?;
+        let port = uri
+            .port_u16()
+            .ok_or_else(|| Error::new("missing MQTT port"))?;
 
         // For MQTT 3 we specify connection version and mode in username field
         // because it doesn't have user properties like MQTT 5.
@@ -268,7 +244,7 @@ impl AgentBuilder {
             .to_owned()
             .unwrap_or_else(|| String::from(""));
 
-        let mut opts = MqttOptions::new(connection.agent_id.to_string(), host, port.as_u16());
+        let mut opts = MqttOptions::new(connection.agent_id.to_string(), host, port);
         opts.set_credentials(username, password);
 
         if let Some(value) = config.clean_session {
@@ -354,7 +330,7 @@ impl Agent {
     }
 
     pub fn id(&self) -> &AgentId {
-        &self.address.id()
+        self.address.id()
     }
 
     /// Publish a message.
@@ -493,10 +469,7 @@ impl Agent {
     where
         S: SubscriptionTopic,
     {
-        let mut topic = subscription.subscription_topic(self.id(), self.address.version())?;
-        if let Some(ref group) = maybe_group {
-            topic = format!("$share/{group}/{topic}", group = group, topic = topic);
-        };
+        let topic = self.get_topic(subscription, maybe_group)?;
 
         self.tx
             .try_send(Request::Subscribe(Subscribe::new(topic, qos)))
@@ -505,6 +478,62 @@ impl Agent {
         Ok(())
     }
 
+    /// Unsubscribe from a topic.
+    ///
+    /// Note that the unsubscribing is actually gets confirmed on receiving
+    /// `AgentNotification::Unsuback` notification.
+    ///
+    /// # Arguments
+    ///
+    /// * `subscription` – the [Subscription](struct.Subscription.html).
+    /// * `maybe_group` – [SharedGroup](struct.SharedGroup.html) in case of multicast subscription.
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// agent.unsubscribe(
+    ///     &Subscription::multicast_requests(Some("v1")),
+    ///     Some(&group),
+    /// )?;
+    ///
+    /// match rx.recv_timeout(Duration::from_secs(5)) {
+    ///     Ok(AgentNotification::Unsuback(_)) => (),
+    ///     Ok(other) => panic!("Expected to receive unsuback notification, got {:?}", other),
+    ///     Err(err) => panic!("Failed to receive unsuback notification: {}", err),
+    /// }
+    /// ```
+    pub fn unsubscribe<S>(
+        &mut self,
+        subscription: &S,
+        maybe_group: Option<&SharedGroup>,
+    ) -> Result<(), Error>
+    where
+        S: SubscriptionTopic,
+    {
+        let topic = self.get_topic(subscription, maybe_group)?;
+
+        self.tx
+            .try_send(Request::Unsubscribe(Unsubscribe::new(topic)))
+            .map_err(|e| Error::new(&format!("error creating MQTT subscription, {}", e)))?;
+
+        Ok(())
+    }
+
+    fn get_topic<S>(
+        &self,
+        subscription: &S,
+        maybe_group: Option<&SharedGroup>,
+    ) -> Result<String, Error>
+    where
+        S: SubscriptionTopic,
+    {
+        let mut topic = subscription.subscription_topic(self.id(), self.address.version())?;
+        if let Some(ref group) = maybe_group {
+            topic = format!("$share/{group}/{topic}", group = group, topic = topic);
+        };
+        Ok(topic)
+    }
+
     #[cfg(feature = "queue-counter")]
     pub fn get_queue_counter(&self) -> QueueCounterHandle {
         self.queue_counter.clone()
@@ -660,7 +689,7 @@ impl ConnectionProperties {
 
 impl Authenticable for ConnectionProperties {
     fn as_account_id(&self) -> &AccountId {
-        &self.agent_id.as_account_id()
+        self.agent_id.as_account_id()
     }
 }
 
@@ -678,7 +707,7 @@ impl Addressable for ConnectionProperties {
 pub enum AgentNotification {
     Message(Result<IncomingMessage<String>, String>, MessageData),
     Reconnection,
-    Disconnection,
+    ConnectionError,
     Puback(PubAck),
     Pubrec(PubRec),
     Pubcomp(PubComp),
